from __future__ import annotations

from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt

from ._list import QCommandList
from ._commands import Command


class QCommandPalette(QtW.QWidget):
    """A Qt command palette widget."""

    def __init__(self, parent: QtW.QWidget = None):
        super().__init__(parent)

        self._line = QtW.QLineEdit()
        self._list = QCommandList()
        _layout = QtW.QVBoxLayout(self)
        _layout.addWidget(self._line)
        _layout.addWidget(self._list)
        self.setLayout(_layout)

        self._line.textChanged.connect(self._on_text_changed)
        self._list.commandClicked.connect(self._on_command_clicked)

    def match_color(self) -> str:
        """The color used for the matched characters."""
        return self._list.match_color()

    def set_match_color(self, color: str):
        """Set the color used for the matched characters."""
        return self._list.set_match_color(color)

    def add_command(self, cmd: Command):
        self._list.add_command(cmd)
        return None

    def extend_command(self, list_of_commands: list[Command]):
        self._list.extend_command(list_of_commands)
        return None

    def install_to(self, parent: QtW.QWidget):
        self.setParent(parent, Qt.WindowType.SubWindow)
        self.hide()

    def _on_text_changed(self, text: str):
        self._list.update_for_text(text)
        return None

    def _on_command_clicked(self, index: int):
        self._list.execute(index)
        self.hide()
        return None

    def focusOutEvent(self, a0: QtGui.QFocusEvent) -> None:
        self.hide()
        return super().focusOutEvent(a0)

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if a0.modifiers() == Qt.KeyboardModifier.NoModifier:
            key = a0.key()
            if key == Qt.Key.Key_Escape:
                self.hide()
            elif key == Qt.Key.Key_Return:
                self.hide()
                self._list.execute()
            elif key == Qt.Key.Key_Up:
                self._list.move_selection(-1)
            elif key == Qt.Key.Key_Down:
                self._list.move_selection(1)
        return super().keyPressEvent(a0)

    def show(self):
        self._line.setText("")
        self._list.update_for_text("")
        super().show()
        if parent := self.parentWidget():
            parent_rect = parent.rect()
            self_size = self.size()
            w, _ = self_size.width(), self_size.height()
            topleft = parent.rect().topLeft()
            topleft.setX(topleft.x() + (parent_rect.width() - w) / 2)
            topleft.setY(topleft.y() + 3)
            self.move(topleft)

        self.raise_()
        self._line.setFocus()
        return None



from __future__ import annotations
from dataclasses import dataclass
from typing import Callable, Generic, TypeVar


_R = TypeVar("_R")


@dataclass
class Command(Generic[_R]):
    """A command representation."""

    function: Callable[..., _R]
    title: str
    desc: str
    tooltip: str = ""

    def __call__(self) -> _R:
        return self.function()

    def fmt(self) -> str:
        """Format command for display in the palette."""
        if self.title:
            return f"{self.title}: {self.desc}"
        return self.desc

    def matches(self, input_text: str) -> bool:
        """Return True if the command matches the input text."""
        fmt = self.fmt().lower()
        words = input_text.lower().split(" ")
        return all(word in fmt for word in words)



from qt_command_palette import get_palette, get_storage

palette = get_palette(name=__name__)
storage = get_storage(name=__name__)


@storage.mark_getter("a")
def get_value():
    return 1


storage.mark_constant("b", 2)


def test_mark_getter():
    assert storage.call(lambda a: a) == 1


def test_mark_constant():
    assert storage.call(lambda b: b) == 2



import sys
from qt_command_palette import get_palette, get_storage


def define(text: str):
    def fn():
        print("called:", text)

    fn.__doc__ = f"Prints {text!r}"
    fn.__name__ = text
    return fn


if __name__ == "__main__":
    from qtpy.QtWidgets import QApplication, QWidget

    app = QApplication([])
    main = QWidget()
    main.setMinimumSize(400, 300)

    palette = get_palette(name=__name__)
    storage = get_storage(name=__name__)

    @storage.mark_getter("widget")
    def get_widget():
        return main

    group_0 = palette.add_group("Example")
    group_1 = palette.add_group("Something new")

    for txt in ["foo", "bar", "baz"]:
        group_0.register(define(txt))

    for txt in ["hello world", "goobye world", "hello again"]:
        group_1.register(define(txt))

    @group_1.register("Print widget")
    def print_widget(widget: QWidget):
        print(widget)

    palette.install(main, "Ctrl+Shift+P")
    main.show()
    sys.exit(app.exec_())



from __future__ import annotations
from typing import Any, TYPE_CHECKING, Iterator
import logging
import re

from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt, Signal

from ._commands import Command

logger = logging.getLogger(__name__)


def colored(text: str, color: str) -> str:
    return f"<b><font color={color!r}>{text}</font></b>"


class QCommandMatchModel(QtCore.QAbstractListModel):
    """A list model for the command palette."""

    def __init__(self, parent: QtW.QWidget = None):
        super().__init__(parent)
        self._commands: list[Command] = []
        self._max_matches = 24

    def rowCount(self, parent: QtCore.QModelIndex = None) -> int:
        return self._max_matches

    def data(self, index: QtCore.QModelIndex, role: int = ...) -> Any:
        return QtCore.QVariant()

    def flags(self, index: QtCore.QModelIndex) -> Qt.ItemFlag:
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable


class QCommandLabel(QtW.QLabel):
    """The label widget to display a command in the palette."""

    def __init__(self, cmd: Command | None = None):
        super().__init__()
        if cmd is not None:
            self.set_command(cmd)
        else:
            self._command_text = ""

    def command(self) -> Command:
        """Command bound to this label."""
        return self._command

    def set_command(self, cmd: Command) -> None:
        """Set command to this widget."""
        command_text = cmd.fmt()
        self._command_text = command_text
        self._command = cmd
        self.setText(command_text)
        self.setToolTip(cmd.tooltip)

    def command_text(self) -> str:
        """The original command text."""
        return self._command_text

    def set_text_colors(self, input_text: str, /, color: str = "blue"):
        """Set label text color based on the input text."""
        if input_text == "":
            return None
        text = self.command_text()
        words = input_text.split(" ")
        pattern = re.compile("|".join(words), re.IGNORECASE)

        output_texts: list[str] = []
        last_end = 0
        for match_obj in pattern.finditer(text):
            output_texts.append(text[last_end : match_obj.start()])
            word = match_obj.group()
            colored_word = colored(word, color)
            output_texts.append(colored_word)
            last_end = match_obj.end()
        output_texts.append(text[last_end:])

        self.setText("".join(output_texts))
        return None


class QCommandList(QtW.QListView):
    commandClicked = Signal(int)  # one of the items is clicked

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setModel(QCommandMatchModel(self))
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.NoSelection)
        self._selected_index = 0
        self._label_widgets: list[QCommandLabel] = []
        self._current_max_index = 0
        for i in range(self.model()._max_matches):
            lw = QCommandLabel()
            self._label_widgets.append(lw)
            self.setIndexWidget(self.model().index(i), lw)
        self.pressed.connect(self._on_clicked)

        self._match_color = "#468cc6"

    def match_color(self) -> str:
        return self._match_color

    def set_match_color(self, color: str):
        self._match_color = color

    def _on_clicked(self, index: QtCore.QModelIndex) -> None:
        if index.isValid():
            self.commandClicked.emit(index.row())
            return None

    def move_selection(self, dx: int) -> None:
        self._selected_index += dx
        self._selected_index = max(0, self._selected_index)
        self._selected_index = min(self._current_max_index - 1, self._selected_index)
        self.update_selection()
        return None

    def update_selection(self):
        index = self.model().index(self._selected_index)
        self.selectionModel().setCurrentIndex(
            index, QtCore.QItemSelectionModel.SelectionFlag.ClearAndSelect
        )
        return None

    @property
    def all_commands(self) -> list[Command]:
        return self.model()._commands

    def add_command(self, command: Command) -> None:
        self.all_commands.append(command)
        return None

    def extend_command(self, commands: list[Command]) -> None:
        """Extend the list of commands."""
        self.all_commands.extend(commands)
        return None

    def command_at(self, index: int) -> Command:
        return self.indexWidget(self.model().index(index)).command()

    def set_command_at(self, index: int, cmd: Command) -> None:
        self.indexWidget(self.model().index(index)).set_command(cmd)
        return None

    def iter_command(self) -> Iterator[Command]:
        for i in range(self.model().rowCount()):
            if not self.isRowHidden(i):
                yield self.command_at(i)

    def execute(self, index: int | None = None) -> None:
        """Execute the currently selected command."""
        if index is None:
            index = self._selected_index
        cmd = self.command_at(index)
        logger.debug(f"executing command: {cmd.fmt()}")
        cmd()
        # move to the top
        self.all_commands.remove(cmd)
        self.all_commands.insert(0, cmd)

    def update_for_text(self, input_text: str) -> None:
        """Update the list to match the input text."""
        self._selected_index = 0
        max_matches = self.model()._max_matches
        row = 0
        for cmd in self.all_commands:
            if cmd.matches(input_text):
                self.setRowHidden(row, False)
                lw = self.indexWidget(self.model().index(row))
                lw.set_command(cmd)
                lw.set_text_colors(input_text, color=self._match_color)
                row += 1

                if row >= max_matches:
                    self._current_max_index = max_matches
                    break
        else:
            self._current_max_index = row
            for row in range(row, max_matches):
                self.setRowHidden(row, True)
        self.update_selection()
        self.update()
        return None

    if TYPE_CHECKING:

        def model(self) -> QCommandMatchModel:
            ...

        def indexWidget(self, index: QtCore.QModelIndex) -> QCommandLabel:
            ...



from __future__ import annotations
from typing import Any, Callable, TypeVar
import inspect

_R = TypeVar("_R")


class Storage:
    """The variable storage."""

    _INSTANCES: dict[str, Storage] = {}

    def __init__(self):
        self._varmap: dict[str, Callable[[], Any]] = {}

    def mark_getter(self, name: str, func: Callable[[], Any] | None = None):
        def wrapper(f: Callable[[], Any]):
            self._varmap[name] = f
            return f

        return wrapper if func is None else wrapper(func)

    def mark_constant(self, name: str, value: Any):
        self._varmap[name] = lambda: value

    @classmethod
    def instance(cls, name: str = "") -> Storage:
        if name not in cls._INSTANCES:
            cls._INSTANCES[name] = Storage()
        return cls._INSTANCES[name]

    def call(self, func: Callable[..., _R]) -> _R:
        """Call a function with variables from the storage."""
        args = []
        for v in inspect.signature(func).parameters.keys():
            if getter := self._varmap.get(v, None):
                args.append(getter())
            else:
                raise ValueError(f"Variable {v} not found in storage")
        return func(*args)


def get_storage(name: str = "") -> Storage:
    """Get the name specific storage instance."""
    return Storage.instance(name)



from qt_command_palette import get_palette

palette = get_palette(name=__name__)


def test_register_no_args():
    group = palette.add_group("test-1")

    @group.register
    def foo():
        pass

    assert palette.commands[-1].desc == "foo"


def test_register_desc():
    group = palette.add_group("test-2")

    @group.register(desc="test func")
    def foo():
        """some tooltip"""

    assert palette.commands[-1].desc == "test func"
    assert palette.commands[-1].tooltip == "some tooltip"


def test_register_tooltip():
    group = palette.add_group("test-3")

    @group.register(tooltip="tooltip")
    def foo():
        pass


def test_register_args():
    group = palette.add_group("test-4")

    @group.register("test func", "tooltip")
    def foo():
        pass


def test_register_with_app_name():
    from qt_command_palette import get_palette

    palette = get_palette("test")

    from . import _file_0, _file_1  # noqa

    assert len(palette.commands) == 2



from __future__ import annotations
from functools import wraps
from typing import Any, Callable, TypeVar, overload, TYPE_CHECKING
import weakref
import inspect
from ._commands import Command
from ._storage import Storage

if TYPE_CHECKING:
    from ._widget import QCommandPalette
    from qtpy import QtWidgets as QtW

_F = TypeVar("_F", bound=Callable)


@inspect.signature
def register_with_func(
    func: Callable,
    title: str | None = None,
    desc: str | None = None,
    tooltip: str | None = None,
):
    """Template function to provide signature to register() with 'func' argument."""


@inspect.signature
def register_without_func(
    title: str | None = None, desc: str | None = None, tooltip: str | None = None
):
    """Template function to provide signature to register() without 'func' argument."""


class CommandPalette:
    """The command palette interface."""

    def __init__(self, name: str) -> None:
        self._commands: list[Command] = []
        self._widget_map: dict[int, QCommandPalette] = {}
        self._name = name

    @property
    def commands(self) -> list[Command]:
        return self._commands.copy()

    @overload
    def register(
        self,
        func: _F,
        title: str | None,
        desc: str | None = None,
        tooltip: str | None = None,
    ) -> _F:
        ...

    @overload
    def register(
        self,
        title: str | None,
        desc: str | None = None,
        tooltip: str | None = None,
    ) -> Callable[[_F], _F]:
        ...

    def register(self, *args, **kwargs):
        """
        Register a function to the command palette.
        """
        if len(args) > 0 and callable(args[0]):
            bound = register_with_func.bind(*args, **kwargs)
        else:
            bound = register_without_func.bind(*args, **kwargs)

        bound.apply_defaults()
        bound_args = bound.arguments
        func = bound_args.pop("func", None)

        # update defaults
        title = bound_args["title"]
        desc = bound_args["desc"]
        tooltip = bound_args["tooltip"]

        if title is None:
            title = ""

        def wrapper(func: _F) -> _F:
            nonlocal title, desc, tooltip
            if desc is None:
                desc = getattr(func, "__name__", repr(func))
            if tooltip is None:
                tooltip = getattr(func, "__doc__", "") or ""

            storage = Storage.instance(self._name)

            @wraps(func)
            def _func():
                return storage.call(func)

            cmd = Command(_func, title, desc, tooltip)
            self._commands.append(cmd)
            return func

        return wrapper if func is None else wrapper(func)

    def add_group(self, title: str) -> CommandGroup:
        return CommandGroup(title, parent=self)

    __default = object()

    def get_widget(self, parent: Any = __default) -> QCommandPalette:
        from ._widget import QCommandPalette

        _id = id(parent)
        if (widget := self._widget_map.get(_id)) is None:
            widget = QCommandPalette()
            widget.extend_command(self._commands)
            self._widget_map[_id] = widget
        return widget

    def show_widget(self, parent: Any = __default) -> None:
        """Show command palette widget."""
        self.get_widget(parent).show()
        return None

    def install(self, parent: QtW.QWidget, keys: str | None = None) -> None:
        """
        Install command palette on a Qt widget.

        Parameters
        ----------
        parent : QtW.QWidget
            The widget to install on.
        keys : str, optional
            If given, this key sequence will be used to show the command palette.
        """
        widget = self.get_widget(parent)
        widget.install_to(parent)
        if keys is not None:
            _register_shortcut(keys, parent, lambda: self.show_widget(parent))
        return None


class CommandGroup:
    def __init__(self, title: str, parent: CommandPalette) -> None:
        self._palette_ref = weakref.ref(parent)
        self._title = title

    def __repr__(self) -> str:
        return f"CommandGroup<{self.title}>"

    @property
    def palette(self) -> CommandPalette:
        """The parent command palette object."""
        if palette := self._palette_ref():
            return palette
        raise RuntimeError("CommandPalette is already destroyed.")

    @property
    def title(self) -> str:
        """The title of this group."""
        return self._title

    @overload
    def register(
        self, func: _F, desc: str | None = None, tooltip: str | None = None
    ) -> _F:
        ...

    @overload
    def register(
        self, desc: str | None = None, tooltip: str | None = None
    ) -> Callable[[_F], _F]:
        ...

    def register(self, *args, **kwargs):
        if "title" in kwargs:
            raise TypeError("register() got an unexpected keyword argument 'title'")
        if len(args) > 0:
            if callable(args[0]):
                args = args[:1] + (self.title,) + args[1:]
            else:
                args = (self.title,) + args
        else:
            args = (self.title,)

        return self.palette.register(*args, **kwargs)


def _register_shortcut(keys: str, parent: QtW.QWidget, target: Callable):
    """Register a callback to a key-binding globally."""
    from qtpy import QT6, QtGui

    if QT6:
        from qtpy.QtGui import QShortcut
    else:
        from qtpy.QtWidgets import QShortcut

    shortcut = QShortcut(QtGui.QKeySequence(keys), parent)
    shortcut.activated.connect(target)
    return None


_GLOBAL_PALETTES: dict[str, CommandPalette] = {}
_DEFAULT_PALETTE = CommandPalette(name="default")


def get_palette(name: str | None = None) -> CommandPalette:
    """
    Get the global command palette object.

    Examples
    --------
    >>> palette = get_palette()  # get the default palette
    >>> palette = get_palette("my_module")  # get a palette for specific app
    """
    global _GLOBAL_PALETTES

    if name is None:
        return _DEFAULT_PALETTE
    if not isinstance(name, str):
        raise TypeError(f"Expected str, got {type(name).__name__}")
    if (palette := _GLOBAL_PALETTES.get(name, None)) is None:
        palette = _GLOBAL_PALETTES[name] = CommandPalette(name=name)
    return palette


def add_group(title: str) -> CommandGroup:
    """
    Add a command group to the global command palette.

    Examples
    --------
    >>> group = add_group("My Commands")
    """
    return get_palette().add_group(title)


@overload
def register(
    func: _F,
    title: str | None,
    desc: str | None = None,
    tooltip: str | None = None,
) -> _F:
    ...


@overload
def register(
    title: str | None,
    desc: str | None = None,
    tooltip: str | None = None,
) -> Callable[[_F], _F]:
    ...


def register(*args, **kwargs):
    """
    Register a function to the global command palette.

    Examples
    --------
    >>> @register
    ... def my_command():
    ...     print("Hello World!")
    >>> @register("My Command")
    ... def my_command():
    ...     print("Hello World!")
    """
    return get_palette().register(*args, **kwargs)
